from collections import deque
from typing import List, Literal

import jsonpatch
from pydantic import PrivateAttr

from rustic_ai.core.guild.agent import Agent, ProcessContext
from rustic_ai.core.guild.agent_ext.depends.llm.models import LLMMessage
from rustic_ai.core.guild.agent_ext.mixins.state_refresher import StateRefresherMixin
from rustic_ai.core.state.models import StateUpdateFormat
from rustic_ai.llm_agent.memories.memories_store import MemoriesStore


class StateBackedMemoriesStore(MemoriesStore):

    memory_type: Literal["state_backed"] = "state_backed"
    memory_size: int = 12

    _memory: deque = PrivateAttr()

    def remember(self, agent: Agent, ctx: ProcessContext, message: LLMMessage) -> None:
        if not isinstance(agent, StateRefresherMixin):
            return

        if not self._memory:
            self._memory = deque(maxlen=self.memory_size)
            state_memories = agent._state.get("memories", [])
            self._memory.extend(state_memories)

        original_memories = {"memories": list(self._memory)}
        self._memory.append(message.model_dump())

        new_memories = {"memories": list(self._memory)}

        update = jsonpatch.make_patch(original_memories, new_memories)

        agent.update_state(
            ctx,
            update_format=StateUpdateFormat.JSON_PATCH,
            update=update,
            update_path="/",
        )

    def recall(self, agent: Agent, ctx: ProcessContext, context: List[LLMMessage]) -> List[LLMMessage]:
        if not isinstance(agent, StateRefresherMixin):
            return []
        self._memory = agent.get_agent_state().get("memories", [])
        return list(self._memory)



from .tools_manager_plugin import (
    ToolsManagerPlugin,
    ToolspecsListProvider,
    ToolspecsProvider,
)
from .tools_provider import ToolsProvider

__all__ = [
    "ToolsManagerPlugin",
    "ToolspecsProvider",
    "ToolspecsListProvider",
    "ToolsProvider",
]



import json
from typing import List, Literal

from rustic_ai.core import Message
from rustic_ai.core.guild.agent import Agent, ProcessContext
from rustic_ai.core.guild.agent_ext.depends.llm.models import (
    ArrayOfContentParts,
    ChatCompletionRequest,
    ChatCompletionResponse,
    LLMMessage,
    TextContentPart,
)
from rustic_ai.core.utils.basic_class_utils import get_qualified_class_name
from rustic_ai.llm_agent.memories.memories_store import MemoriesStore


class HistoryBasedMemoriesStore(MemoriesStore):
    memory_type: Literal["history_based"] = "history_based"
    text_only: bool = True

    def remember(self, agent: Agent, ctx: ProcessContext, message: LLMMessage) -> None:
        # No-op: History-based memory does not store messages explicitly.
        pass

    def _filter_text_messages(self, messages):
        text_messages = []
        for msg in messages:
            if isinstance(msg.content, str):
                text_messages.append(msg)
            elif isinstance(msg.content, ArrayOfContentParts):
                is_text = False
                for part in msg.content.root:
                    if isinstance(part, TextContentPart):
                        is_text = True
                    else:
                        msg.content.root.remove(part)

                    if is_text:
                        text_messages.append(msg)
        messages = text_messages
        return messages

    def _deduplicate_messages(self, messages):
        seen_content = set()
        unique_messages = []

        for msg in messages:
            # Create a stable key for possibly unhashable content (e.g., dict, complex types)
            try:
                key = msg.content if hash(msg.content) is not None else msg.content
            except TypeError:
                try:
                    key = json.dumps(
                        msg.content,
                        default=lambda o: getattr(o, "model_dump", lambda: repr(o))(),
                        sort_keys=True,
                    )
                except TypeError:
                    key = repr(msg.content)

            if key not in seen_content:
                seen_content.add(key)
                unique_messages.append(msg)
        return unique_messages

    def _extract_messages_from_history(self, enriched_history):
        messages: List[LLMMessage] = []

        for entry in enriched_history:
            message = Message.from_json(entry)
            if message.format == get_qualified_class_name(ChatCompletionRequest):
                payload = ChatCompletionRequest.model_validate(message.payload)
                messages.extend(payload.messages)
            elif message.format == get_qualified_class_name(ChatCompletionResponse):
                # Implement retrieval logic for ChatCompletionResponse messages
                payload = ChatCompletionResponse.model_validate(message.payload)
                messages.append(payload.choices[0].message)
        return messages

    def recall(self, agent: Agent, ctx: ProcessContext, context: List[LLMMessage]) -> List[LLMMessage]:
        enriched_history = ctx.get_context().get("enriched_history", [])

        messages = self._extract_messages_from_history(enriched_history)

        # Deduplicate messages by their content

        unique_messages = self._deduplicate_messages(messages)

        messages = unique_messages

        if self.text_only:
            messages = self._filter_text_messages(messages)

        return messages



from typing import Optional

from rustic_ai.core.guild.agent import Agent, ProcessContext, processor
from rustic_ai.core.guild.agent_ext.depends.llm.llm import LLM
from rustic_ai.core.guild.agent_ext.depends.llm.models import (
    ChatCompletionRequest,
    SystemMessage,
)
from rustic_ai.core.messaging.core.message import Message
from rustic_ai.core.utils.json_utils import JsonDict
from rustic_ai.llm_agent.llm_agent_conf import LLMAgentConfig
from rustic_ai.llm_agent.llm_agent_helper import LLMAgentHelper
from rustic_ai.llm_agent.llm_agent_utils import LLMAgentUtils


class LLMAgent(Agent[LLMAgentConfig]):
    """
    A simple LLM Agent that simply invokes an LLM.
    """

    _system_prompt: Optional[str] = None
    _updated: bool = False

    @processor(
        ChatCompletionRequest,
        predicate=lambda self, msg: LLMAgentUtils.has_no_attachments(msg),
        depends_on=["llm"],
    )
    def invoke_llm(self, ctx: ProcessContext[ChatCompletionRequest], llm: LLM):
        """
        Invoke the LLM with the given context. All the LLM configuration parameters are passed along.
        """
        prompt = ctx.payload

        if self._system_prompt:
            # If the system prompt was updated, add it to the messages.
            messages = [SystemMessage(content=self._system_prompt)] + prompt.messages
            prompt = prompt.model_copy(update={"messages": messages})
        elif self.config.default_system_prompt:
            # If there is a default system prompt, use it.
            messages = [SystemMessage(content=self.config.default_system_prompt)] + prompt.messages
            prompt = prompt.model_copy(update={"messages": messages})

        LLMAgentHelper.invoke_llm_and_handle_response(
            self,
            self.config,
            llm,
            ctx,
            prompt,
        )

    def _is_system_prompt_update(self, msg: Message) -> bool:
        config = self.config
        if config.system_prompt_generator and msg.format == config.system_prompt_generator.update_on_message_format:
            return True
        return False

    @processor(JsonDict, predicate=lambda self, msg: self._is_system_prompt_update(msg))
    def update_system_prompt(self, ctx: ProcessContext[JsonDict]):
        """
        Update the system prompt using the system prompt generator.
        """
        if not self.config.system_prompt_generator:
            return

        self._system_prompt = (
            self.config.system_prompt_generator.generate_prompt(ctx.payload)
            if self.config.system_prompt_generator
            else self.config.default_system_prompt
        )
        self._updated = True



from abc import ABC, abstractmethod
from typing import List, Optional

from pydantic import BaseModel, Field, model_validator

from rustic_ai.core.guild.agent import Agent, ProcessContext
from rustic_ai.core.guild.agent_ext.depends.llm.llm import LLM
from rustic_ai.core.guild.agent_ext.depends.llm.models import (
    ChatCompletionRequest,
    ChatCompletionResponse,
)


class LLMCallWrapper(BaseModel, ABC):
    """
    Base class for request wrap processors (plugins inherit from this).
    The implementations of this class will process the prompts before sending them to the LLM
    and also get the response after receiving it from the LLM.
    """

    kind: Optional[str] = Field(default=None, frozen=True, description="FQCN of the wrap processor class")

    @model_validator(mode="after")
    def _enforce_kind_matches_class(self):
        fqcn = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
        if self.kind and self.kind != fqcn:
            raise ValueError(f"`kind` must be {fqcn!r}, got {self.kind!r}")

        if not self.kind:
            object.__setattr__(self, "kind", fqcn)

        return self

    @abstractmethod
    def preprocess(
        self,
        agent: Agent,
        ctx: ProcessContext[ChatCompletionRequest],
        request: ChatCompletionRequest,
        llm: LLM,
    ) -> ChatCompletionRequest:
        """
        Preprocess the prompt before sending it to the LLM.
        This method can modify the prompt as needed.
        """
        pass

    @abstractmethod
    def postprocess(
        self,
        agent: Agent,
        ctx: ProcessContext[ChatCompletionRequest],
        final_prompt: ChatCompletionRequest,
        llm_response: ChatCompletionResponse,
        llm: LLM,
    ) -> Optional[List[BaseModel]]:
        """
        Postprocess the response received from the LLM.
        This method can perform an action using the response.
        Any values returned from this method will be sent as messages.
        """
        pass



from rustic_ai.core.guild.agent import Agent, ProcessContext
from rustic_ai.core.guild.agent_ext.depends.llm.llm import LLM
from rustic_ai.core.guild.agent_ext.depends.llm.models import (
    ChatCompletionRequest,
    ChatCompletionTool,
)
from rustic_ai.llm_agent.plugins.request_preprocessor import RequestPreprocessor


class ToolsProvider(RequestPreprocessor):
    """
    A request preprocessor that adds a list of tools to the request.
    If the request already has tools, the provider's tools are appended to the existing list.
    If the request has no tools, the provider's tools become the request's tools.
    If the LLM makes a tool call, it will be returned in the ChatCompletetionResponse.
    """

    tools: list[ChatCompletionTool]

    def preprocess(
        self,
        agent: Agent,
        ctx: ProcessContext[ChatCompletionRequest],
        request: ChatCompletionRequest,
        llm: LLM,
    ) -> ChatCompletionRequest:

        _tools = list(request.tools or [])
        _tools.extend(self.tools)

        return request.model_copy(update={"tools": _tools})



import logging
from typing import List, Optional, Union

import openai

from rustic_ai.core.guild.agent import Agent, ProcessContext
from rustic_ai.core.guild.agent_ext.depends.llm.llm import LLM
from rustic_ai.core.guild.agent_ext.depends.llm.models import (
    ChatCompletionError,
    ChatCompletionRequest,
    ChatCompletionResponse,
    LLMMessage,
    ResponseCodes,
)
from rustic_ai.llm_agent.llm_agent_conf import LLMAgentConfig
from rustic_ai.llm_agent.plugins.llm_call_wrapper import LLMCallWrapper
from rustic_ai.llm_agent.plugins.response_postprocessor import ResponsePostprocessor


class LLMAgentHelper:

    @staticmethod
    def invoke_llm_completion(
        agent: Agent,
        config: LLMAgentConfig,
        llm: LLM,
        ctx: ProcessContext[ChatCompletionRequest],
        prompt: ChatCompletionRequest,
    ) -> Union[ChatCompletionResponse, ChatCompletionError]:
        """
        Invoke the LLM completion with the given context.
        The fields from the chat completion request, Agent Config, and the LLM are combined.
        The LLM Configuration is used as the base, overwritten by Agent config and then
        the Chat Completion Request.
        """

        pre_processors = config.request_preprocessors or []
        wrap_processors = config.llm_request_wrappers or []
        post_processors = config.response_postprocessors or []

        for pre_processor in pre_processors:
            prompt = pre_processor.preprocess(agent=agent, ctx=ctx, request=prompt, llm=llm)

        for wrap_processor in wrap_processors:
            prompt = wrap_processor.preprocess(agent=agent, ctx=ctx, request=prompt, llm=llm)

        config_dict = config.get_llm_params()
        prompt_dict = prompt.model_dump(exclude_none=True)

        final_prompt = {
            **config_dict,
            **prompt_dict,
        }

        cc_request: ChatCompletionRequest = ChatCompletionRequest.model_validate(final_prompt)

        max_retries = config.max_retries

        return LLMAgentHelper._invoke_llm_completion(
            agent=agent,
            ctx=ctx,
            llm=llm,
            cc_request=cc_request,
            wrap_processors=wrap_processors,
            post_processors=post_processors,
            max_retries=max_retries,
            model=config.model,
        )

    @staticmethod
    def _invoke_llm_completion(
        agent: Agent,
        ctx: ProcessContext[ChatCompletionRequest],
        llm: LLM,
        cc_request: ChatCompletionRequest,
        wrap_processors: List[LLMCallWrapper],
        post_processors: List[ResponsePostprocessor],
        max_retries: int,
        model: Optional[str] = None,
    ) -> Union[ChatCompletionResponse, ChatCompletionError]:

        response = llm.completion(cc_request, model)

        new_messages = []

        try:
            for wrap_processor in reversed(wrap_processors):
                post_msg = wrap_processor.postprocess(
                    agent=agent,
                    ctx=ctx,
                    final_prompt=cc_request,
                    llm_response=response,
                    llm=llm,
                )
                if post_msg:
                    new_messages.extend(post_msg)

            for post_processor in post_processors:
                post_msg = post_processor.postprocess(
                    agent=agent,
                    ctx=ctx,
                    final_prompt=cc_request,
                    llm_response=response,
                    llm=llm,
                )
                if post_msg:
                    new_messages.extend(post_msg)

            reasoning = response.choices[0].message.reasoning_content if response.choices else ""

            for msg in new_messages:
                if reasoning:
                    ctx.send(msg, reason=reasoning)
                else:
                    ctx.send(msg)

        except Exception as e:  # pragma: no cover
            logging.error(f"Error in post processing: {e}")
            if max_retries > 0:
                logging.info(f"Retrying LLM call, remaining retries: {max_retries}")
                return LLMAgentHelper._invoke_llm_completion(
                    agent=agent,
                    ctx=ctx,
                    llm=llm,
                    cc_request=cc_request,
                    wrap_processors=wrap_processors,
                    post_processors=post_processors,
                    max_retries=max_retries - 1,
                )
            else:
                return ChatCompletionError(
                    status_code=ResponseCodes.RESPONSE_PROCESSING_ERROR,
                    message=str(e),
                    model=agent.name,
                    request_messages=ctx.payload.messages,
                )

        return response

    @staticmethod
    def process_api_status_error(
        model_name: str,
        ctx: ProcessContext[ChatCompletionRequest],
        status_code: ResponseCodes,
        error: openai.APIStatusError,
        messages: List[Union[LLMMessage]],
    ) -> ChatCompletionError:  # pragma: no cover
        """
        Process API status error and return a ChatCompletionError object.
        """
        cc_error = ChatCompletionError(
            status_code=status_code,
            message=error.message,
            response=error.response.text if error.response else None,
            model=model_name,
            request_messages=messages,
            body=error.body if hasattr(error, "body") else None,
        )

        return cc_error

    @staticmethod
    def invoke_llm_and_handle_response(
        agent: Agent,
        config: LLMAgentConfig,
        llm: LLM,
        ctx: ProcessContext[ChatCompletionRequest],
        prompt: ChatCompletionRequest,
    ) -> ChatCompletionResponse | ChatCompletionError:
        """
        Invoke the LLM and handle the response.
        The fields from the chat completion request, Agent Config, and the LLM are combined.
        The LLM Configuration is used as the base, overwritten by Agent config and then
        the Chat Completion Request.
        """
        try:
            chat_response = LLMAgentHelper.invoke_llm_completion(agent, config, llm, ctx, prompt)

            if isinstance(chat_response, ChatCompletionError):
                ctx.send_error(chat_response)
                return chat_response

            reasoning = chat_response.choices[0].message.reasoning_content if chat_response.choices else ""

            if reasoning:
                ctx.send(chat_response, reason=reasoning)
            else:
                ctx.send(chat_response)
            return chat_response
        except openai.APIStatusError as e:  # pragma: no cover
            logging.error(f"Error in LLM completion: {e}")
            # Publish the error message
            error = LLMAgentHelper.process_api_status_error(
                model_name=config.model,
                ctx=ctx,
                status_code=ResponseCodes(e.status_code),
                error=e,
                messages=ctx.payload.messages,
            )

            ctx.send_error(error)
            return error
        except Exception as e:  # pragma: no cover
            logging.error(f"Unexpected error in LLM completion: {e}")
            error = ChatCompletionError(
                status_code=ResponseCodes.INTERNAL_SERVER_ERROR,
                message=str(e),
                model=agent.name,
                request_messages=ctx.payload.messages,
            )

            ctx.send_error(error)
            return error



from .llm_agent import LLMAgent
from .llm_agent_conf import LLMAgentConfig
from .llm_agent_utils import LLMAgentUtils
from .memories import (
    HistoryBasedMemoriesStore,
    MemoriesStore,
    QueueBasedMemoriesStore,
    StateBackedMemoriesStore,
)
from .plugins import (
    LLMCallWrapper,
    PromptGenerator,
    RequestPreprocessor,
    ResponsePostprocessor,
    TemplatedPromptGenerator,
)

__all__ = [
    "LLMAgent",
    "LLMAgentConfig",
    "LLMAgentUtils",
    "PromptGenerator",
    "TemplatedPromptGenerator",
    "MemoriesStore",
    "QueueBasedMemoriesStore",
    "HistoryBasedMemoriesStore",
    "StateBackedMemoriesStore",
    "LLMCallWrapper",
    "RequestPreprocessor",
    "ResponsePostprocessor",
    "TemplatedPromptGenerator",
]



from .llm_call_wrapper import LLMCallWrapper
from .prompt_generators import PromptGenerator, TemplatedPromptGenerator
from .request_preprocessor import RequestPreprocessor
from .response_postprocessor import ResponsePostprocessor

__all__ = [
    "LLMCallWrapper",
    "RequestPreprocessor",
    "ResponsePostprocessor",
    "PromptGenerator",
    "TemplatedPromptGenerator",
]



from enum import Enum
from typing import Any, List, Optional, Type, TypeVar, Union, cast

from pydantic import BaseModel, ConfigDict, Field, field_serializer, field_validator
from typing_extensions import Annotated

from rustic_ai.core.guild.dsl import BaseAgentProps
from rustic_ai.core.utils.basic_class_utils import get_class_from_name
from rustic_ai.llm_agent.plugins.llm_call_wrapper import LLMCallWrapper
from rustic_ai.llm_agent.plugins.prompt_generators import PromptGenerator
from rustic_ai.llm_agent.plugins.request_preprocessor import RequestPreprocessor
from rustic_ai.llm_agent.plugins.response_postprocessor import ResponsePostprocessor


class Models(str, Enum):
    gpt_5 = "gpt-5"
    gpt_5_chat = "gpt-5-chat"
    gpt_5_mini = "gpt-5-mini"
    gpt_5_nano = "gpt-5-nano"

    gpt_4o = "gpt-4o"
    gpt_4o_mini = "gpt-4o-mini"

    gpt_4_1 = "gpt-4.1"
    gpt_4_1_mini = "gpt-4.1-mini"
    gpt_4_1_nano = "gpt-4.1-nano"

    o3 = "o3"
    o3_pro = "o3-pro"
    o3_deep_research = "o3-deep-research"
    o3_mini = "o3-mini"

    o4_mini = "o4-mini"

    gpt_4 = "gpt-4"
    gpt_4_turbo = "gpt-4-turbo"

    gemini_pro = "gemini-pro"
    gemini_2_5_pro = "gemini-2.5-pro"
    gemini_flash = "gemini-flash"
    gemini_2_5_flash = "gemini-2.5-flash"
    gemini_2_5_flash_lite = "gemini-2.5-flash-lite"

    gemini_2_0_flash = "gemini-2.0-flash"
    gemini_2_0_flash_lite = "gemini-2.0-flash-lite"

    gemini_1_5_pro = "gemini-1.5-pro"
    gemini_1_5_flash = "gemini-1.5-flash"

    claude_opus_4_1 = "claude-opus-4-1"
    claude_opus_4_0 = "claude-opus-4-0"
    claude_sonnet_4_0 = "claude-sonnet-4-0"
    claude_sonnet_3_5 = "claude-3-5-sonnet-latest"
    claude_haiku_3_5 = "claude-3-5-haiku-latest"


T = TypeVar("T", bound=BaseModel)


def _build_plugins(value: Any, base_type: type[T]) -> List[T]:
    """
    Accepts:
      - list of dicts/instances (mixed allowed)
      - single dict
      - single instance

    Dict shape: {"kind": "pkg.mod.Class", ...kwargs}
    Instances are kept as-is.
    """
    if value is None:
        return []

    items = value if isinstance(value, list) else [value]
    out: List[T] = []

    for i, item in enumerate(items):
        # Already constructed instance (programmatic usage)
        if isinstance(item, base_type):
            # mypy: item is a runtime instance of base_type (a subclass of BaseModel)
            out.append(cast(T, item))
            continue

        # Dict spec with FQCN
        if isinstance(item, dict):
            cls_path = item.get("kind")
            if not cls_path or not isinstance(cls_path, str):
                raise TypeError(f"Plugin spec at index {i} must include string 'kind' (FQCN)")
            cls: Type[Any] = get_class_from_name(cls_path)
            if not issubclass(cls, base_type):
                raise TypeError(f"{cls_path!r} is not a subclass of {base_type.__name__}")
            kwargs = {k: v for k, v in item.items() if k != "kind"}
            instance = cls(**kwargs)
            out.append(cast(T, instance))
            continue

        raise TypeError(
            f"Unsupported plugin spec at index {i}: expected dict or {base_type.__name__} instance; "
            f"got {type(item).__name__}"
        )

    return out


class LLMAgentConfig(BaseAgentProps):
    """
    Base configuration for LLM agents.
    """

    model_config = ConfigDict(extra="ignore")

    model: Annotated[Union[str, Models], Field(examples=["gpt-5"])]
    """
    ID of the model to use. See the [model endpoint compatibility](/docs/models/model-endpoint-compatibility) table
    for details on which models work with the Chat API.
    """

    base_url: Optional[str] = None
    """Base URL for the LLM API."""

    api_version: Optional[str] = None
    """Version for the LLM API."""

    custom_llm_provider: Optional[str] = None
    """Custom LLM provider to use, e.g., 'bedrock' for Amazon Bedrock or 'vertex' for Google Vertex AI."""

    timeout: Optional[float] = None
    """
    Timeout for the LLM API requests.
    If the completion takes longer than this time, the request will be aborted.
    """

    max_retries: int = 0
    """
    Maximum number of retries for LLM requests if post processing fails.
    This is useful if the LLM returns a response that can not be processed successfully
    e.g. due to a parsing error in tools response.
    """

    default_system_prompt: Optional[str] = None
    """
    Default system prompt to use if no other mechanism updates the prompt.
    """

    system_prompt_generator: Optional[PromptGenerator] = Field(discriminator="type", default=None)
    """
    Mechanism to update the system prompt based on messages from other agents messages.
    """

    request_preprocessors: List[RequestPreprocessor] = Field(default_factory=list)
    """
    0 or more request preprocessors to apply before sending prompts to the LLM.
    The order of preprocessors matters: they are applied in the order they are listed.
    """

    llm_request_wrappers: List[LLMCallWrapper] = Field(default_factory=list)
    """
    0 or more request wrap processors to apply before sending prompts to the LLM
    and after receiving the response from the LLM. The plugin may modify the prompts before sending them to the LLM
    and also act on the response after receiving it from the LLM.
    The order of wrap processors matters: they are applied in the order they are listed.
    """

    response_postprocessors: List[ResponsePostprocessor] = Field(default_factory=list)
    """
    0 or more response postprocessors to apply after receiving the response from the LLM.
    All the post processors get the response as is and can act on it.
    Any modifications to the response don't have any affect on other post processors or the final response.
    """

    send_response: bool = True
    """
    Whether to send the original LLM response back to the conversation.
    If False, only the post-processed messages (if any) will be sent back.
    This is useful if the LLM response is only used for tool calls and not for direct responses.
    """

    vertex_location: Optional[str] = None
    """
    Place where vertex model is deployed (us-central1, asia-southeast1, etc.). If None, attempts to use the VERTEXAI_LOCATION environment variable.
    """

    vertex_project: Optional[str] = None
    """
    The Google Cloud project ID. If None, attempts to use the VERTEXAI_PROJECT environment variable.
    """

    @field_validator("request_preprocessors", mode="before")
    @classmethod
    def _coerce_req(cls, v):
        return _build_plugins(v, RequestPreprocessor)

    @field_validator("llm_request_wrappers", mode="before")
    @classmethod
    def _coerce_wrap(cls, v):
        return _build_plugins(v, LLMCallWrapper)

    @field_validator("response_postprocessors", mode="before")
    @classmethod
    def _coerce_resp(cls, v):
        return _build_plugins(v, ResponsePostprocessor)

    @field_serializer("request_preprocessors", mode="plain")
    def _serialize_preprocessors(self, preprocessors):
        """Ensure subclass fields are preserved during serialization."""
        return [p.model_dump(exclude_none=True) for p in preprocessors]

    @field_serializer("llm_request_wrappers", mode="plain")
    def _serialize_wrappers(self, wrappers):
        """Ensure subclass fields are preserved during serialization."""
        return [w.model_dump(exclude_none=True) for w in wrappers]

    @field_serializer("response_postprocessors", mode="plain")
    def _serialize_postprocessors(self, postprocessors):
        """Ensure subclass fields are preserved during serialization."""
        return [p.model_dump(exclude_none=True) for p in postprocessors]

    _non_llm_fields = {
        "max_retries",
        "request_preprocessors",
        "llm_request_wrappers",
        "response_postprocessors",
        "default_system_prompt",
        "system_prompt_generator",
        "send_response",
    }

    def get_llm_params(self) -> dict:
        """
        Get the LLM parameters from the config, excluding non-LLM fields.
        """
        return self.model_dump(exclude={*self._non_llm_fields})



from abc import ABC, abstractmethod
from functools import cached_property
import json
from typing import List, Optional

from pydantic import (
    BaseModel,
    Field,
    field_serializer,
    field_validator,
    model_validator,
)

from rustic_ai.core.guild.agent_ext.depends.llm.models import (
    ChatCompletionRequest,
    ChatCompletionResponse,
    ChatCompletionTool,
    SystemMessage,
)
from rustic_ai.core.guild.agent_ext.depends.llm.tools_manager import ToolSpec
from rustic_ai.core.utils.basic_class_utils import get_class_from_name
from rustic_ai.llm_agent.plugins.llm_call_wrapper import LLMCallWrapper


class ToolspecsProvider(BaseModel, ABC):

    kind: Optional[str] = Field(default=None, frozen=True, description="FQCN of the wrap processor class")

    additional_system_prompt: Optional[str] = None

    def model_post_init(self, __context) -> None:
        if not self.kind:
            object.__setattr__(self, "kind", f"{self.__class__.__module__}.{self.__class__.__qualname__}")

    @model_validator(mode="after")
    def _enforce_kind_matches_class(self):
        fqcn = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
        if self.kind and self.kind != fqcn:
            raise ValueError(f"`kind` must be {fqcn!r}, got {self.kind!r}")
        return self

    @abstractmethod
    def get_toolspecs(self) -> List[ToolSpec]:
        pass

    @cached_property
    def toolspecs_by_name(self) -> dict[str, ToolSpec]:
        return {tool.name: tool for tool in self.get_toolspecs()}

    def get_toolspec_by_name(self, name: str) -> Optional[ToolSpec]:
        return self.toolspecs_by_name.get(name, None)

    @cached_property
    def chat_tools(self) -> List[ChatCompletionTool]:
        return [tool.chat_tool for tool in self.get_toolspecs()]


class ToolspecsListProvider(ToolspecsProvider):

    tools: List[ToolSpec] = Field(description="List of ToolSpec objects", min_length=1)

    def get_toolspecs(self) -> List[ToolSpec]:
        return self.tools


class ToolsHelper:

    @staticmethod
    def parse_tool_args(provider: ToolspecsProvider, name: str, args: dict) -> Optional[BaseModel]:
        """
        Parses the arguments for a tool by its name and validates them against the parameter class.
        This method attempts to create an instance of the parameter class using the provided arguments.

        :param tool: The ToolSpec object representing the tool.
        :param name: The name of the tool.
        :param args: A dictionary of arguments to be validated.
        :return: An instance of the parameter class if validation is successful, None otherwise.
        """

        tool = provider.get_toolspec_by_name(name)
        if not tool:
            return None

        return tool.parse_args(args)

    @staticmethod
    def extract_tool_calls(provider: ToolspecsProvider, response: ChatCompletionResponse) -> List[BaseModel]:
        """
        Extract tool calls from the model's response.

        :param response: The ChatCompletionResponse object containing the model's response.
        :return: A list of tool calls extracted from the response.
        """

        tool_calls: List[BaseModel] = []
        for choice in response.choices:
            if choice.message and choice.message.tool_calls:
                calls = choice.message.tool_calls
                for call in calls:
                    name = call.function.name
                    args = json.loads(call.function.arguments)
                    tool = ToolsHelper.parse_tool_args(provider, name, args)

                    if tool:
                        tool_calls.append(tool)

        return tool_calls


class ToolsManagerPlugin(LLMCallWrapper):

    toolset: ToolspecsProvider

    def preprocess(
        self,
        agent,
        ctx,
        request,
        llm,
    ) -> ChatCompletionRequest:
        messages = request.messages
        if messages is None:
            messages = []

        if self.toolset.additional_system_prompt:
            if len(messages) > 0 and messages[0].role == "system":
                # If system prompt already exists, add the additional prompt as a new system message after it.
                messages = [messages[0]] + [SystemMessage(content=self.toolset.additional_system_prompt)] + messages[1:]
            else:
                messages = [SystemMessage(content=self.toolset.additional_system_prompt)] + messages

        tools = self.toolset.chat_tools

        tools = request.tools + tools if request.tools else tools

        request = request.model_copy(update={"tools": tools, "messages": messages})
        return request

    def postprocess(
        self,
        agent,
        ctx,
        final_prompt,
        llm_response,
        llm,
    ) -> Optional[List[BaseModel]]:
        tool_calls = ToolsHelper.extract_tool_calls(self.toolset, llm_response)
        return tool_calls

    @field_validator("toolset", mode="before")
    @classmethod
    def _load_toolset(cls, v):
        if isinstance(v, dict):
            kind = v.get("kind")
            if not kind:
                # Default to list provider if kind is not specified
                toolset_cls = ToolspecsListProvider
            else:
                toolset_cls = get_class_from_name(kind)

            if not issubclass(toolset_cls, ToolspecsProvider):
                raise ValueError(f"Toolset class {toolset_cls} is not a subclass of ToolspecsProvider")
            return toolset_cls.model_validate(v)

        elif isinstance(v, ToolspecsProvider):
            return v

        else:
            raise ValueError("Toolset must be a dict or an instance of ToolspecsProvider")

    @field_serializer("toolset", mode="plain")
    def _serialize_toolset(self, toolset):
        return toolset.model_dump()



from abc import ABC, abstractmethod
from typing import Literal

from pydantic import BaseModel

from rustic_ai.core.utils.json_utils import JsonDict


class PromptGenerator(BaseModel, ABC):
    """Base class for prompt generators."""

    type: Literal["template"]
    update_on_message_format: str

    @abstractmethod
    def generate_prompt(self, data: JsonDict) -> str:
        """Generate a prompt based on the provided data."""
        pass


class TemplatedPromptGenerator(PromptGenerator):
    """Prompt generator that uses a template for prompts."""

    type: Literal["template"] = "template"
    update_on_message_format: str

    template: str

    def generate_prompt(self, dict: JsonDict) -> str:
        return self.template.format(**dict)



from abc import ABC, abstractmethod
from typing import Optional

from pydantic import BaseModel, Field, model_validator

from rustic_ai.core.guild.agent import Agent, ProcessContext
from rustic_ai.core.guild.agent_ext.depends.llm.llm import LLM
from rustic_ai.core.guild.agent_ext.depends.llm.models import ChatCompletionRequest


class RequestPreprocessor(BaseModel, ABC):
    """
    Base class for request preprocessors (plugins inherit from this).
    The implementations of this class will process the prompts before sending them to the LLM.
    """

    kind: Optional[str] = Field(default=None, frozen=True, description="FQCN of the preprocessor class")

    @model_validator(mode="after")
    def _enforce_kind_matches_class(self):
        fqcn = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
        if self.kind and self.kind != fqcn:
            raise ValueError(f"`kind` must be {fqcn!r}, got {self.kind!r}")

        if not self.kind:
            object.__setattr__(self, "kind", fqcn)

        return self

    @abstractmethod
    def preprocess(
        self,
        agent: Agent,
        ctx: ProcessContext[ChatCompletionRequest],
        request: ChatCompletionRequest,
        llm: LLM,
    ) -> ChatCompletionRequest:
        """
        Preprocess the prompt before sending it to the LLM.
        This method can modify the prompt as needed.
        """
        pass
